#!/usr/bin/env python                    #this enables a user to run the file by typing only it's name (no need for python prefix)

"""    block comment
Created 20251202

@author: dturney
"""


# TA_data are assumed to be 2D matrices with wavelengths along the 1st row, and with probe delay times along the 1st column

import sys
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy.interpolate import griddata

# PyQt5 Imports (The GUI Library)
from PyQt5.QtWidgets import (QApplication, QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QGridLayout, QLineEdit, QLabel, QPushButton, QStatusBar, QGroupBox)
from PyQt5.QtCore import QThread, pyqtSignal, Qt
from scipy.interpolate import CubicSpline
from scipy.interpolate import griddata

# Matplotlib Imports (The Plotting Library)
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT 
from matplotlib.figure import Figure

# Import the shared functions and classes
import shared_functions_classes as TA_sh


# TA_data are assumed to be 2D matrices with wavelengths along the 1st row, and with probe delay times along the 1st column


class TA_t0_correction_and_background_removal_GUI():   
    def __init__(self, TA_matrix_input):
        self.app = QApplication(sys.argv)                                           # Create the QApplication instance           
        self.TA_matrix_input = TA_matrix_input
        self.main_pyqt5_window = TA_t0_correction_and_background_removal_main_Window(self.TA_matrix_input)   # Create the main pyqt5 window
        self.main_pyqt5_window.show()                                               # Display the window
        sys.exit(self.app.exec())                                                   # Start the event loop        



class TA_t0_correction_and_background_removal_main_Window(QMainWindow):    #This nomenclature causes this class to inherit all the methods and properties of the QMainWindow class from PyQT5.
    def __init__(self, TA_matrix_input):
        
        super().__init__() #This executes a lot of code that builds the PyQT5 window using the QMainWindow class.
        
        # pyqt5 window setup: needs the QMainWindow object to be inherited into this class. 
        self.setWindowTitle("TA t0 Correction and Background Removal")
        self.dpi = QApplication.primaryScreen().logicalDotsPerInch() #Get screen DPI
        self.window_width_inches = 15
        self.window_height_inches = 10.5
        self.resize(int(self.window_width_inches * self.dpi), int(self.window_height_inches * self.dpi))  # Width, Height in pixels (converted from inches)

        # Create the main pyqt5 object (Qwidget) and then create the geometric layout that will hold a LHS box (for the matplotlib stuff) and RHS box (for the pyqt5 Controls).
        self.main_window_widget = QWidget()
        self.setCentralWidget(self.main_window_widget)
        self.horiz_pyqt5_layout = QHBoxLayout(self.main_window_widget) # Horizontal: Plot on Left, Controls on Right

        # Add the vertical pyqt5 layout to the horizontal (main) layout. Create the geometric layout of the LHS box of the QHBoxLayout window: to hold the matplotlib toolbar on top and the matplotlib figure below. This ensures the toolbar sits right on top of the plot
        self.vertical_layout_widget = QWidget()
        self.vert_pyqt5_layout = QVBoxLayout(self.vertical_layout_widget)   # Vertical pyqt5 layout: toolbar on top, canvas below
        self.vert_pyqt5_layout.setContentsMargins(0, 0, 0, 0) # Remove extra spacing
        self.horiz_pyqt5_layout.addWidget(self.vertical_layout_widget, stretch=4)

        # Create the Control Panel (aka the pyqt5 layout for inserting buttons and textboxes) 
        self.control_panel = QWidget()
        self.control_panel_layout = QVBoxLayout(self.control_panel)
        self.control_panel_layout.setAlignment(Qt.AlignTop)
        self.horiz_pyqt5_layout.addWidget(self.control_panel, stretch=1) # Add the control panel to main horizontal layout (since it's added to the main horizontal layout 1st, it's located on the right side).

        # Create matplotlib Figure and PyQT5 Canvas (actually, it's a pyqt5 imitation of a matplotlib toolbar). 
        self.fig_han = Figure(dpi=self.dpi)   # We don't need to specify figsize here since the canvas will auto-scale to fill the pyqt5 layout space.
        self.canvas = FigureCanvasQTAgg(self.fig_han)
        self.toolbar = Custom_QT5_Toolbar(self.canvas, self)
        self.toolbar.setStyleSheet("QLabel { font-family: 'Courier New', Consolas, monospace; font-size: 13pt; white-space: pre; }") # We have to use specific fonts to get the toolbar message to not wiggle with different number font widths.
        self.vert_pyqt5_layout.addWidget(self.toolbar)
        self.vert_pyqt5_layout.addWidget(self.canvas, stretch=4) # Add Canvas to the main horizontal layout (Takes up most of the space. Since it's added to the main horizontal laout 2nd, it's on the left side).

        # Load the TA matrix data
        self.TA_matrix_input = TA_matrix_input
        self.TA_matrix = TA_sh.get_TA_matrix(self.TA_matrix_input)
        # Extract the wavelengths and time delays
        self.TA_matrix_wavelengths = self.TA_matrix[0,1:]
        self.TA_matrix_delay_times = self.TA_matrix[1:,0]
        # Crop the TA image down to remove the wavelengths and delay times
        self.TA_data = self.TA_matrix[1:,1:]
        self.TA_data_rows, self.TA_data_cols = self.TA_data.shape
        # Create interpolator for the status bar readout
        self.TA_matrix_interpolator = RegularGridInterpolator((self.TA_matrix_delay_times, np.flip(self.TA_matrix_wavelengths)), np.fliplr(self.TA_data), bounds_error=False, fill_value=None)

        # Create the Matplotlib Figure and Canvas
        self.pcolormesh_transects = TA_sh.embedded_matplotlib_pcolormesh_transects(parent_pyqt5_QMainWindow=self, frac_horiz=1.0, frac_vert=1.0)

        # Add simple instructions
        self.fig_han.text(0.76, 0.068, 'Instructions:')
        self.fig_han.text(0.76, 0.03, 'Double mouse click (left button)\nto move crosshairs.')

        # create pyqt5 controls for t0 correction and background removal
        t0_grp_box = QGroupBox("t0 Correction Controls")
        t0_grp_box.setStyleSheet('QLabel { font-size: 10pt; } QLineEdit { font-size: 10pt; } QPushButton { font-size: 10pt; }')
        t0_grid_layout = QGridLayout()
        t0_grid_layout.setVerticalSpacing(2)
        t0_grid_layout.setContentsMargins(2, 5, 2, 2)
        self.btn_add_t0 = QPushButton("Add point for t0 fit")
        self.btn_add_t0.clicked.connect(self.add_t0_point)
        t0_grid_layout.addWidget(self.btn_add_t0, 0, 0)
        # Button: Fit t0 line
        self.btn_fit_t0 = QPushButton("Fit t0 Line")
        self.btn_fit_t0.clicked.connect(self.fit_t0)
        t0_grid_layout.addWidget(self.btn_fit_t0, 1, 0)
        # Button: Correct t0 & Background
        self.btn_correct = QPushButton("Correct t0 & Background")
        self.btn_correct.clicked.connect(self.correct_t0_background)
        t0_grid_layout.addWidget(self.btn_correct, 2, 0)
        # Button: Save Corrected Matrix
        self.btn_save_matrix = QPushButton("Save Corrected Matrix")
        self.btn_save_matrix.clicked.connect(self.save_TA_matrix)
        t0_grid_layout.addWidget(self.btn_save_matrix)
        t0_grp_box.setLayout(t0_grid_layout)
        self.control_panel_layout.addWidget(t0_grp_box)

        # Initialize t0 correction variables
        # Create a list to hold a history of clicked points
        self.click_history = [[-10000,-10000]]    # start with a dummy click to allow yourself to check if values are increasing, then cut away this zero for final fitting
        self.t0_each_wavelength = np.zeros(len(self.TA_matrix_wavelengths))

        # Create plot objects for t0 visualization (Line for fit, Scatter for points)
        # We access the axis handle that resides inside the pcolormesh_transects object
        self.t0_points_scatter = self.pcolormesh_transects.TA_image_axis_han.scatter([], [], color='white', s=30, edgecolor='k', zorder=10)
        self.t0_fit_line, = self.pcolormesh_transects.TA_image_axis_han.plot([], [], linewidth=0.7, alpha=0.7, color='k', zorder=11)
        self.t0_fit_line.set_visible(False)
        self.t0_fit_line.set_linewidth(1.0)        

        # Set focus so keyboard works immediately
        self.canvas.setFocusPolicy(Qt.ClickFocus)
        self.canvas.setFocus()

        # Initial Draw
        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)

        # Connect Matplotlib Events
        self.canvas.mpl_connect('button_press_event', self.on_TA_matrix_mouse_click)
        self.canvas.mpl_connect('key_press_event', self.on_TA_matrix_up_down_left_right)

    ############################################
    ####### Callbacks for Interactivity with the PyQt5 Control Buttons and Textboxes

    def on_TA_matrix_up_down_left_right(self, event):
        ### Handles arrow keys.
        # Check boundaries
        if event.key == 'up' and self.pcolormesh_transects.crosshair_h_idx < self.TA_data_rows - 1:
            self.pcolormesh_transects.crosshair_h_idx += 1
        elif event.key == 'down' and self.pcolormesh_transects.crosshair_h_idx > 0:
            self.pcolormesh_transects.crosshair_h_idx -= 1
        elif event.key == 'left' and self.pcolormesh_transects.crosshair_v_idx < self.TA_data_cols - 1:
            self.pcolormesh_transects.crosshair_v_idx += 1
        elif event.key == 'right' and self.pcolormesh_transects.crosshair_v_idx > 0:
            self.pcolormesh_transects.crosshair_v_idx -= 1
        else:
            return # Ignore other keys
        self.pcolormesh_transects.update_crosshair_lines()
        self.pcolormesh_transects.update_transects()


    def on_TA_matrix_mouse_click(self, event):
        ### Handles mouse clicks to move crosshairs.
        if event.inaxes != self.pcolormesh_transects.TA_image_axis_han: return
        # We only care about Double Clicks (per your original script)
        if event.dblclick:
            # Update indices
            self.pcolormesh_transects.crosshair_v_idx = int(np.abs(self.TA_matrix_wavelengths - event.xdata).argmin())
            self.pcolormesh_transects.crosshair_h_idx = int(np.abs(self.TA_matrix_delay_times - event.ydata).argmin())
            self.pcolormesh_transects.update_crosshair_lines()
            self.pcolormesh_transects.update_transects()
            if (event.button == 3):
                self.add_t0_point()


    def add_t0_point(self):
            #Adds the current crosshair location to the list of points for t0 fitting.
            # Get current indices from the transect object
            w_idx = self.pcolormesh_transects.crosshair_v_idx
            t_idx = self.pcolormesh_transects.crosshair_h_idx
            
            # Get actual values
            w_val = self.TA_matrix_wavelengths[w_idx]
            t_val = self.TA_matrix_delay_times[t_idx]
            
            # Add to history
            self.click_history.append([w_val, t_val])
            
            # Update the visual scatter plot of selected points
            self.t0_fit_line.set_visible(True)
            self.t0_points_scatter.set_offsets(np.array(self.click_history))
            self.canvas.draw()
            
            self.status_bar.showMessage(f"Added t0 point: {w_val:.1f} nm, {t_val:.2f} ps")


    def fit_t0(self):
        #Fits a cubic spline to the selected t0 points.
        click_history = np.array(self.click_history[1:])     # cut away the first click_history data because it was artificial
        if len(click_history) < 2: # Need more than 2 for the fit
            self.status_bar.showMessage("Error: Select at least 2 points to fit t0.") 
            return
        # Sort points by wavelength (required for spline)
        click_history = click_history[click_history[:, 0].argsort()]
        try:
            # Create spline fit
            t0_spline_interpolator = CubicSpline(click_history[:, 0], click_history[:, 1], extrapolate=True)
            self.t0_each_wavelength[:] = t0_spline_interpolator(self.TA_matrix_wavelengths)
            # Update the fit line on the plot
            self.t0_fit_line.set_data(self.TA_matrix_wavelengths, self.t0_each_wavelength)
            self.t0_fit_line.set_visible(True)
            self.canvas.draw()
            self.status_bar.showMessage("t0 fit updated.")
        except Exception as e:
            self.status_bar.showMessage(f"Fit failed: {e}")


    def correct_t0_background(self):
        #Performs the t0 chirp correction, then background removal.
        if np.all(self.t0_each_wavelength == 0):
            self.status_bar.showMessage("Error: Please fit t0 first.")
            return
        self.status_bar.showMessage("Processing... t0 correction will take a minute or two. Be patient.")
        self.thread = QThread()              # Create a QThread object
        self.worker = pyqt5_thread_worker_t0_correction(parent_pyqt5_QMainWindow=self)    # Create the Thread Worker
        self.worker.moveToThread(self.thread) # Move the worker to the thread
        self.thread.started.connect(self.worker.run)        # Connect Signals: When thread starts -> Run worker logic
        self.worker.finished.connect(self.thread.quit)          # Connect Signals: When worker finishes -> Quit thread, Clean up objects
        self.worker.finished.connect(self.worker.deleteLater)   # Connect Signals: When worker finishes -> Quit thread, Clean up objects
        self.thread.finished.connect(self.thread.deleteLater)   # Connect Signals: When worker finishes -> Quit thread, Clean up objects
        self.worker.result_ready.connect(self.on_t0_correction_complete)   # Connect Signals: When data is ready -> Update the GUI
        self.worker.error_occurred.connect(lambda e: self.status_bar.showMessage(f"Error: {e}"))    # Connect Signals: if error occurs -> Show error message        
        self.btn_correct.setEnabled(False); self.btn_save_matrix.setEnabled(False); self.btn_add_t0.setEnabled(False); self.btn_fit_t0.setEnabled(False)  # Disable the buttons so they don't click it twice 
        self.thread.start()             # Start the thread


    def on_t0_correction_complete(self, corrected_data):
        # Destroy the fit line and reset click_history
        self.click_history = [[-10000,-10000]]    # start with a value of zero in there to allow yourself to check if values are increasing, then cut away this zero for final fitting
        self.t0_points_scatter.set_offsets(np.array(self.click_history))
        self.t0_fit_line.set_visible(False)
        self.t0_fit_line.set_data([],[])
        
        # Count NaNs, report them, fill them
        nan_count = np.sum(np.isnan(corrected_data))
        if nan_count > 0:
            print(f"{nan_count} NaNs found and set to 0.")
            corrected_data[np.isnan(corrected_data)] = 0

        self.TA_data = corrected_data   # Update the TA_matrix Data
        self.pcolormesh_transects.TA_image_axis_han.collections[0].set_array(self.TA_data.ravel())      # Update the TA_matrix plot data
        self.TA_matrix_interpolator = RegularGridInterpolator((self.TA_matrix_delay_times, np.flip(self.TA_matrix_wavelengths)), np.fliplr(self.TA_data), bounds_error=False, fill_value=None) # Recreate the interpolator
        self.pcolormesh_transects.update_transects()
        self.canvas.draw()
        
        # Re-enable buttons
        self.btn_correct.setEnabled(True); self.btn_save_matrix.setEnabled(True); self.btn_add_t0.setEnabled(True); self.btn_fit_t0.setEnabled(True)  # Re-allow these buttons
        self.status_bar.showMessage("Correction Complete via Background Thread.")



    def save_TA_matrix(self):
        """Saves the current (corrected) TA matrix to disk."""
        if isinstance(self.TA_matrix_input, str):
            filename = f"{self.TA_matrix_input}.t0_corr.csv"
        else:
            filename = "output.t0_corr.csv"

        # Reconstruct the full matrix structure (Wavelengths in row 0, Delays in col 0)
        output_matrix = np.zeros(self.TA_matrix.shape)
        output_matrix[0, 1:] = self.TA_matrix_wavelengths
        output_matrix[1:, 0] = self.TA_matrix_delay_times
        output_matrix[1:, 1:] = self.TA_data

        try:
            np.savetxt(filename, output_matrix, delimiter=',')
            self.status_bar.showMessage(f"Saved Matrix: {filename}")
        except Exception as e:
            self.status_bar.showMessage(f"Save Failed: {e}")






class pyqt5_thread_worker_t0_correction(QThread):
    finished = pyqtSignal()                      # define a signal: Task is done
    result_ready = pyqtSignal(object)            # define a signal: Here is the data (sends a numpy array)
    error_occurred = pyqtSignal(str)             # define a signal: Something went wrong

    def __init__(self, parent_pyqt5_QMainWindow):
        super().__init__()
        self.parent_pyqt5_QMainWindow = parent_pyqt5_QMainWindow

    def run(self):
        try:
            # Perform heavy calculation here: t0 CHIRP CORRECTION (Interpolation)
            delay_times = self.parent_pyqt5_QMainWindow.TA_matrix_delay_times
            wavelengths = self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths
            t0_each_wavelength = self.parent_pyqt5_QMainWindow.t0_each_wavelength
            # Create grid of original data points
            grid_delay_times = np.outer(delay_times, np.ones(len(wavelengths)))      #outer product to make a grid
            grid_wavelengths = np.outer(np.ones(len(delay_times)), wavelengths)      #outer product to make a grid
            # Pad the TA_data to avoid NaNs at edges during interpolation
            TA_data_padded = np.vstack((self.parent_pyqt5_QMainWindow.TA_data[0, :], self.parent_pyqt5_QMainWindow.TA_data[0, :], self.parent_pyqt5_QMainWindow.TA_data[0, :], self.parent_pyqt5_QMainWindow.TA_data))
            TA_data_padded = np.vstack((TA_data_padded, TA_data_padded[-1, :], TA_data_padded[-1, :], TA_data_padded[-1, :]))
            top_diff = grid_delay_times[0, :] - grid_delay_times[1, :]; bottom_diff = grid_delay_times[-1, :] - grid_delay_times[-2, :] 
            grid_delay_times_padded = np.vstack((grid_delay_times[0, :]+top_diff*3, grid_delay_times[0, :]+top_diff*2, grid_delay_times[0, :]+top_diff, grid_delay_times ) )
            grid_delay_times_padded = np.vstack((grid_delay_times_padded, grid_delay_times_padded[-1, :]+bottom_diff, grid_delay_times_padded[-1, :]+bottom_diff*2, grid_delay_times_padded[-1, :]+bottom_diff*3 ))
            grid_wavelengths_padded = np.vstack((grid_wavelengths[0, :], grid_wavelengths[0, :], grid_wavelengths[0, :], grid_wavelengths))
            grid_wavelengths_padded = np.vstack((grid_wavelengths_padded, grid_wavelengths_padded[-1, :], grid_wavelengths_padded[-1, :], grid_wavelengths_padded[-1, :]))
            t0_corrected_array_delay_times_padded = grid_delay_times_padded - t0_each_wavelength
            # Perform the interpolation
            corrected_data = griddata( (t0_corrected_array_delay_times_padded.ravel(), grid_wavelengths_padded.ravel()), TA_data_padded.ravel(), (grid_delay_times, grid_wavelengths), method='linear')
            # BACKGROUND REMOVAL
            background_each_wavelength = np.zeros(len(wavelengths))
            t0_idx = int(np.abs(delay_times).argmin()) # Index of time zero
            # Logic to determine start index for background averaging
            st_idx = int(delay_times.argmin()) # Find which ench of the TA matrix holds the smallest delay times
            # Iterative background removal
            for repeat in range(0, 30):
                for j in range(1, len(self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths) - 2):
                    if st_idx < t0_idx:
                        background_each_wavelength[j] = np.sum(corrected_data[st_idx:t0_idx-15, j-1:j+1]) / (t0_idx - st_idx) / 3
                        background_each_wavelength[0] = np.sum(corrected_data[st_idx:t0_idx-15, 0]) / (t0_idx - st_idx)
                        background_each_wavelength[-1] = np.sum(corrected_data[st_idx:t0_idx-15, -1]) / (t0_idx - st_idx)
                    elif st_idx > t0_idx:
                        background_each_wavelength[j] = np.sum(corrected_data[t0_idx:st_idx+15, j-1:j+1]) / (st_idx - t0_idx) / 3
                        background_each_wavelength[0] = np.sum(corrected_data[t0_idx:st_idx+15, 0]) / (st_idx - t0_idx)
                        background_each_wavelength[-1] = np.sum(corrected_data[t0_idx:st_idx+15, -1]) / (st_idx - t0_idx)
                corrected_data[:] = corrected_data - background_each_wavelength


            self.result_ready.emit(corrected_data)  # FINISHED: Send the corrected data back to the main thread
            self.finished.emit()                     # # FINISHED: Let the thread know we're done

        except Exception as e:
            self.error_occurred.emit(str(e))






class Custom_QT5_Toolbar(NavigationToolbar2QT):
    def __init__(self, canvas, parent_pyqt5_QMainWindow, coordinates=True):
        super().__init__(canvas, parent_pyqt5_QMainWindow, coordinates)
        self.parent_pyqt5_QMainWindow = parent_pyqt5_QMainWindow
        
    def home(self, *args, **kwargs):
        # Call the Original Home Behavior. This performs the actual view reset in Matplotlib
        super().home(*args, **kwargs)
        # Now I can force my own view
        self.parent_pyqt5_QMainWindow.pcolormesh_transects.TA_image_axis_han.set_xlim(self.parent_pyqt5_QMainWindow.pcolormesh_transects.home_xlim)
        self.parent_pyqt5_QMainWindow.pcolormesh_transects.TA_image_axis_han.set_ylim(self.parent_pyqt5_QMainWindow.pcolormesh_transects.home_ylim)
        self.parent_pyqt5_QMainWindow.canvas.draw()




# Define a hook to ensure exceptions are caught by the debugger/terminal
def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    print("Uncaught exception", file=sys.stderr)
    sys.__excepthook__(exc_type, exc_value, exc_traceback)

sys.excepthook = handle_exception


# if __name__ == "__main__":       # This line is asking: "Is this file the one that started the process?"  We can have only one QApplication instance per process, so we need to ensure this is the main file.
#     # Check if a filename was provided in the command line arguments
#     if len(sys.argv) > 1:
#         TA_matrix_input = sys.argv[1]
#     else:
#         # Fallback default if no file is provided
#         TA_matrix_input = "HHHF_Zn_heme_ZnCl_p425nm_red_300uW.h5"  # default input for testing:
#     # Start the application
#     window = TA_t0_correction_and_background_removal_GUI(TA_matrix_input) #The QApplication instance must be created before this.  The init method recognizes QApplication already exists and uses it.
    



#!/usr/bin/env python                    #this enables a user to run the file by typing only it's name (no need for python prefix)

"""    block comment
Created 20251202

@author: dturney
"""


# TA_data are assumed to be 2D matrices with wavelengths along the 1st row, and with probe delay times along the 1st column

import sys
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy.interpolate import CubicSpline

# PyQt5 Imports (The GUI Library)
from PyQt5.QtWidgets import (QApplication, QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QGridLayout, QLineEdit, QLabel, QPushButton, QStatusBar, QGroupBox, QComboBox)
from PyQt5.QtCore import Qt

# Matplotlib Imports (The Plotting Library)
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT 
from matplotlib.figure import Figure

# Import the shared functions and classes
import shared_functions_classes as TA_sh


# TA_data are assumed to be 2D matrices with wavelengths along the 1st row, and with probe delay times along the 1st column


class TA_merge_matrix_GUI():   
    def __init__(self, TA_matrix_input_filenames):
        self.app = QApplication(sys.argv)                                           # Create the QApplication instance           
        self.TA_matrix_input_filenames = TA_matrix_input_filenames
        self.main_pyqt5_window = TA_merge_matrix_GUI_main_Window(self.TA_matrix_input_filenames)   # Create the main pyqt5 window
        self.main_pyqt5_window.show()                                               # Display the window
        sys.exit(self.app.exec())                                                   # Start the event loop        



class TA_merge_matrix_GUI_main_Window(QMainWindow):    #This nomenclature causes this class to inherit all the methods and properties of the QMainWindow class from PyQT5.
    def __init__(self, TA_matrix_input_filenames):
        print('NOTE: This function always merges TA data into the format of the first TA_matrix in the input list.')

        super().__init__() #This executes a lot of code that builds the PyQT5 window using the QMainWindow class.
        
        self.TA_matrix_input_filenames = TA_matrix_input_filenames

        # pyqt5 window setup: needs the QMainWindow object to be inherited into this class. 
        self.setWindowTitle("TA matrix merger")
        self.dpi = QApplication.primaryScreen().logicalDotsPerInch() #Get screen DPI
        window_taller_factor = 1.23
        self.window_width_inches = 15
        self.window_height_inches = 10.6 * window_taller_factor
        self.resize(int(self.window_width_inches * self.dpi), int(self.window_height_inches * self.dpi))  # Width, Height in pixels (converted from inches)

        # Create the main pyqt5 object (Qwidget) and then create the geometric layout that will hold a LHS box (for the matplotlib stuff) and RHS box (for the pyqt5 Controls).
        self.main_window_widget = QWidget()
        self.setCentralWidget(self.main_window_widget)
        self.horiz_pyqt5_layout = QHBoxLayout(self.main_window_widget) # Horizontal: Plot on Left, Controls on Right

        # Add the vertical pyqt5 layout to the horizontal (i.e. the main) layout. Create the geometric layout of the LHS box of the QHBoxLayout window: to hold the matplotlib toolbar on top and the matplotlib figure below. This ensures the toolbar sits right on top of the plot
        self.vertical_layout_widget = QWidget()
        self.vert_pyqt5_layout = QVBoxLayout(self.vertical_layout_widget)   # Vertical pyqt5 layout: toolbar on top, canvas below
        self.vert_pyqt5_layout.setContentsMargins(0, 0, 0, 0)               # Remove extra spacing
        self.horiz_pyqt5_layout.addWidget(self.vertical_layout_widget, stretch=4)

        # Create the Control Panel (aka the pyqt5 layout for inserting buttons and textboxes) 
        self.control_panel = QWidget()
        self.control_panel_layout = QVBoxLayout(self.control_panel)
        self.control_panel_layout.setAlignment(Qt.AlignTop)
        self.horiz_pyqt5_layout.addWidget(self.control_panel, stretch=1) # Add the control panel to main horizontal layout (since it's added to the main horizontal layout 1st, it's located on the right side).

        # Create matplotlib Figure and PyQT5 Canvas (actually, it's a pyqt5 imitation of a matplotlib toolbar). 
        self.fig_han = Figure(dpi=self.dpi)                 # We don't need to specify figsize here since the canvas will auto-scale to fill the pyqt5 layout space.
        self.canvas = FigureCanvasQTAgg(self.fig_han)
        self.toolbar = Custom_QT5_Toolbar(self.canvas, self)
        self.toolbar.setStyleSheet("QLabel { font-family: 'Courier New', Consolas, monospace; font-size: 13pt; white-space: pre; }") # We have to use specific fonts to get the toolbar message to not wiggle with different number font widths.
        self.vert_pyqt5_layout.addWidget(self.toolbar)
        self.vert_pyqt5_layout.addWidget(self.canvas, stretch=4) # Add Canvas to the main horizontal layout (Takes up most of the space. Since it's added to the main horizontal laout 2nd, it's on the left side).

        # Load the TA matrix data and TA non-pumped probe counts 
        self.TA_matrix = TA_sh.get_TA_matrix(TA_matrix_input_filenames[0])    # Loads the TA matrix data from file
        self.presently_selected_idx = 0                                                         # Initialize Data Variables
        self.TA_matrix_wavelengths = self.TA_matrix[0,1:]                                       # Extract the wavelengths and time delays
        self.TA_matrix_delay_times = self.TA_matrix[1:,0]
        self.TA_data = self.TA_matrix[1:,1:]                                                    # Crop the TA image down to remove the wavelengths and delay times
        self.TA_data_rows, self.TA_data_cols = self.TA_data.shape
        self.TA_matrix_interpolator = RegularGridInterpolator((self.TA_matrix_delay_times, np.flip(self.TA_matrix_wavelengths)), np.fliplr(self.TA_data), bounds_error=False, fill_value=None) # Create interpolator for the status bar readout

        # Create the Matplotlib Figure and Canvas
        self.pcolormesh_transects = TA_sh.embedded_matplotlib_pcolormesh_transects(parent_pyqt5_QMainWindow=self, frac_horiz=1.0, frac_vert=1/window_taller_factor)

        # Add simple instructions
        self.fig_han.text(0.79, 0.068, 'Instructions:')
        self.fig_han.text(0.79, 0.03, 'Double mouse click (left button)\nto move crosshairs.')

        # Group Box to Choose the TA_matrix for merge inspection
        choose_TA_matrix_grp = QGroupBox("Merge & Probe Controls")
        choose_TA_matrix_grp.setStyleSheet('QLabel { font-size: 10pt; } QLineEdit { font-size: 10pt; } QPushButton { font-size: 10pt; } QComboBox { font-size: 10pt; }')
        choose_TA_matrix = QGridLayout()
        choose_TA_matrix.setVerticalSpacing(4)
        choose_TA_matrix.setContentsMargins(2, 5, 2, 2)
        choose_TA_matrix.addWidget(QLabel("Select TA dataset:"), 0, 0)          # TA Dataset Dropdown list
        self.combo_datasets = QComboBox()
        self.combo_datasets.addItems(self.TA_matrix_input_filenames)
        self.combo_datasets.currentIndexChanged.connect(self.update_TA_matrix_plot_via_dropdown)
        self.combo_datasets.setFixedWidth(150)
        choose_TA_matrix.addWidget(self.combo_datasets, 1, 0)
        choose_TA_matrix_grp.setLayout(choose_TA_matrix)
        self.control_panel_layout.addWidget(choose_TA_matrix_grp)

         # Group Box of Buttons/Textboxes for the merge
        merge_grp = QGroupBox("Merge & Probe Controls")
        merge_grp.setStyleSheet('QLabel { font-size: 10pt; } QLineEdit { font-size: 10pt; } QPushButton { font-size: 10pt; } QComboBox { font-size: 10pt; }')
        merge_layout = QGridLayout()
        merge_layout.setVerticalSpacing(4)
        merge_layout.setContentsMargins(2, 5, 2, 2)       
        merge_layout.addWidget(QLabel("Ignore range min (nm):"), 2, 0)              # Zero Out Probe Counts textbox
        self.txt_zero_range_min = QLineEdit("0.0")
        merge_layout.addWidget(self.txt_zero_range_min, 2, 1)
        merge_layout.addWidget(QLabel("Ignore range max (nm):"), 3, 0)              # Zero Out Probe Counts textbox
        self.txt_zero_range_max = QLineEdit("0.0")
        merge_layout.addWidget(self.txt_zero_range_max, 3, 1)
        self.btn_zero_counts = QPushButton("Zero the Ignore Range")               # Zero Out Counts Button
        self.btn_zero_counts.clicked.connect(self.zero_probe_counts)
        merge_layout.addWidget(self.btn_zero_counts, 4, 0, 1, 2)            # row,   column wthin row,   # of rows to occupy,    # of columns to occupy)
        self.btn_smooth = QPushButton("Smooth Merge Fraction")              # Smooth Merge Fraction Button
        self.btn_smooth.clicked.connect(self.smooth_the_merge)              # row,   column wthin row,   # of rows to occupy,    # of columns to occupy)
        merge_layout.addWidget(self.btn_smooth, 5, 0, 1, 2)                 # row,   column wthin row,   # of rows to occupy,    # of columns to occupy)
        self.btn_merge = QPushButton("Merge TA Datasets")                   # Merge Button
        self.btn_merge.clicked.connect(self.merge_TA_data)
        merge_layout.addWidget(self.btn_merge, 6, 0, 1, 2)                  # row,   column wthin row,   # of rows to occupy,    # of columns to occupy)
        merge_grp.setLayout(merge_layout)
        self.control_panel_layout.addWidget(merge_grp)
        
        # Status Bar
        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)

        # Load probe counts for each experiment. Calculate probe fractions for merging
        self.all_probe_counts = np.zeros((len(self.TA_matrix_input_filenames), len(self.TA_matrix_wavelengths)))  # Initialize storage
        print('WARNING: The minimum probe count will be subtracted from each probe spectrum to clear background noise.')
        for i, filename in enumerate(self.TA_matrix_input_filenames):
            raw_wavelength = np.flip(TA_sh.get_TA_matrix(filename)[0,1:])                                           # Load raw data
            raw_probe_counts = np.flip(TA_sh.get_TA_probe_counts(filename))
            raw_probe_count_interpolator = CubicSpline(raw_wavelength, raw_probe_counts, extrapolate=True)          # Interpolate to match the master wavelength axis
            self.all_probe_counts[i,:] = raw_probe_count_interpolator(self.TA_matrix_wavelengths)
            self.all_probe_counts[i,:] = self.all_probe_counts[i,:] - np.min(self.all_probe_counts[i,:])            # Subtract minimum to clear background noise
        self.sqrt_all_probe_counts = np.sqrt(self.all_probe_counts)
        self.fraction_in_merge_each_probe_spectrum = self.sqrt_all_probe_counts.copy()
        sum_sqrt = np.sum(self.sqrt_all_probe_counts, axis=0)                                                       # Calculate sum of square roots across all datasets
        sum_sqrt[sum_sqrt == 0] = 1e-9                                                                              # Prevent divide by zero
        for i in range(len(self.TA_matrix_input_filenames)):
            self.fraction_in_merge_each_probe_spectrum[i,:] = self.sqrt_all_probe_counts[i,:] / sum_sqrt
        
        # Initial plot of the probe counts
        self.probe_counts = self.all_probe_counts[self.presently_selected_idx, :]
        self.ax_probe_counts = self.fig_han.add_axes([0.06, 0.02, 0.66, 0.161], sharex=self.pcolormesh_transects.TA_image_axis_han) # Create axes below the main image. [left bottom width height]
        self.ax_probe_counts.axes.yaxis.set_label_text('probe counts', color='b')
        self.ax_probe_counts.tick_params(axis='y', colors='blue')
        self.line_probe_counts, = self.ax_probe_counts.plot(self.TA_matrix_wavelengths, self.probe_counts, color='b')
        self.ax_probe_counts.set_ylim(0, np.max(self.probe_counts) * 1.1)
        idx = self.pcolormesh_transects.crosshair_v_idx                                                             # Red dot for crosshair tracking.  Note: We access crosshair indices from the shared pcolormesh object
        self.red_dot_probe_counts = self.ax_probe_counts.scatter(self.TA_matrix_wavelengths[idx], self.probe_counts[idx], color='red', s=40, zorder=2)
        self.ax_fraction_in_merge = self.ax_probe_counts.twinx()                                                     # Create a second y-axis that shares the same x-axis
        self.line_fraction_in_merge, = self.ax_fraction_in_merge.plot(self.TA_matrix_wavelengths, self.fraction_in_merge_each_probe_spectrum[self.presently_selected_idx,:], color='k', linestyle=':')
        self.ax_fraction_in_merge.set_ylabel('fraction in merge')
        self.ax_fraction_in_merge.set_ylim(0, 1)
        self.ax_probe_counts.tick_params(labelbottom=False)
        self.ax_probe_counts.grid(True, linestyle=':')
        self.status_bar.showMessage(f"Loaded: {TA_matrix_input_filenames[0]}")

        # Set Focus
        self.canvas.setFocusPolicy(Qt.ClickFocus)
        self.canvas.setFocus()

        # Initial Draw
        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)

        # Connect Matplotlib Events
        self.canvas.mpl_connect('button_press_event', self.on_TA_matrix_mouse_click)
        self.canvas.mpl_connect('key_press_event', self.on_TA_matrix_up_down_left_right)


    def update_probe_count_plot(self):                                                                                  # Updates the secondary plots for probe counts and merge fractions.
            self.probe_counts = self.all_probe_counts[self.presently_selected_idx, :]
            self.line_probe_counts.set_ydata(self.probe_counts)
            self.red_dot_probe_counts.set_offsets([self.TA_matrix_wavelengths[self.pcolormesh_transects.crosshair_v_idx], self.probe_counts[self.pcolormesh_transects.crosshair_v_idx]])
            self.line_fraction_in_merge.set_ydata(self.fraction_in_merge_each_probe_spectrum[self.presently_selected_idx,:])
            self.canvas.draw()


    ############################################
    ####### Callbacks for Interactivity with the PyQt5 Control Buttons and Textboxes


    def on_TA_matrix_up_down_left_right(self, event):
        ### Handles arrow keys.
        # Check boundaries
        if event.key == 'up' and self.pcolormesh_transects.crosshair_h_idx < self.TA_data_rows - 1:
            self.pcolormesh_transects.crosshair_h_idx += 1
        elif event.key == 'down' and self.pcolormesh_transects.crosshair_h_idx > 0:
            self.pcolormesh_transects.crosshair_h_idx -= 1
        elif event.key == 'left' and self.pcolormesh_transects.crosshair_v_idx < self.TA_data_cols - 1:
            self.pcolormesh_transects.crosshair_v_idx += 1
        elif event.key == 'right' and self.pcolormesh_transects.crosshair_v_idx > 0:
            self.pcolormesh_transects.crosshair_v_idx -= 1
        else:
            return # Ignore other keys
        self.pcolormesh_transects.update_crosshair_lines()
        self.pcolormesh_transects.update_transects()
        self.red_dot_probe_counts.set_offsets([self.TA_matrix_wavelengths[self.pcolormesh_transects.crosshair_v_idx], self.probe_counts[self.pcolormesh_transects.crosshair_v_idx]])


    def on_TA_matrix_mouse_click(self, event):
        ### Handles mouse clicks to move crosshairs.
        if event.inaxes != self.pcolormesh_transects.TA_image_axis_han: return
        # We only care about Double Clicks (per your original script)
        if event.dblclick:
            # Update indices
            self.pcolormesh_transects.crosshair_v_idx = int(np.abs(self.TA_matrix_wavelengths - event.xdata).argmin())
            self.pcolormesh_transects.crosshair_h_idx = int(np.abs(self.TA_matrix_delay_times - event.ydata).argmin())
            self.pcolormesh_transects.update_crosshair_lines()
            self.pcolormesh_transects.update_transects()
            self.red_dot_probe_counts.set_offsets([self.TA_matrix_wavelengths[self.pcolormesh_transects.crosshair_v_idx], self.probe_counts[self.pcolormesh_transects.crosshair_v_idx]])
            self.canvas.draw()



    def update_TA_matrix_plot_via_dropdown(self):
        self.presently_selected_idx = self.combo_datasets.currentIndex()                                       # Update the presently selected index
        filename = self.TA_matrix_input_filenames[self.presently_selected_idx]
        self.TA_matrix = TA_sh.get_TA_matrix(filename)                                                          # Reload the TA matrix data from file
        self.TA_data = self.TA_matrix[1:,1:]
        self.TA_matrix_wavelengths = self.TA_matrix[0,1:]                                                       # Extract the wavelengths and time delays
        self.TA_matrix_delay_times = self.TA_matrix[1:,0]
        self.pcolormesh_transects.pclrmsh.set_array(self.TA_data.ravel())                                         # Update the Pcolormesh Data
        self.TA_matrix_interpolator = RegularGridInterpolator((self.TA_matrix_delay_times, np.flip(self.TA_matrix_wavelengths)), np.fliplr(self.TA_data), bounds_error=False, fill_value=None) # Re-create Interpolator
        self.update_probe_count_plot()
        self.pcolormesh_transects.update_transects()
        self.canvas.draw()

        

    def zero_probe_counts(self):                                                                                        #Callback for Zero Out Counts Button. Zeroes out probe counts in the range specified by the QLineEdit.
        try:
            w_range = [float(self.txt_zero_range_min.text()) , float(self.txt_zero_range_max.text())]
            if len(w_range) != 2 or w_range[1] <= w_range[0]:
                self.status_bar.showMessage("Error: Range must be 'min,max' with max > min")
                return
            w_idxs = [np.abs(self.TA_matrix_wavelengths - w_range[0]).argmin() , np.abs(self.TA_matrix_wavelengths - w_range[1]).argmin()]    # Find indices
            self.all_probe_counts[self.presently_selected_idx, np.min(w_idxs):np.max(w_idxs)] = 1e-10                   # Zero out the probe counts in the specified range
            self.sqrt_all_probe_counts[:] = np.sqrt(self.all_probe_counts)                                              # Recalculate square roots
            sum_sqrt = np.sum(self.sqrt_all_probe_counts, axis=0)
            sum_sqrt[sum_sqrt == 0] = 1e-9
            for i in range(len(self.TA_matrix_input_filenames)):
                self.fraction_in_merge_each_probe_spectrum[i,:] = self.sqrt_all_probe_counts[i,:] / sum_sqrt
            self.update_TA_matrix_plot_via_dropdown()                                        # Update the plots
            self.update_probe_count_plot()
            self.status_bar.showMessage(f"Zeroed probe counts between {w_range[0]} and {w_range[1]} nm")
        except ValueError:
            self.status_bar.showMessage("Error: Invalid input. Use format '450,500'")



    def smooth_the_merge(self):                                                                                             #Callback for Smooth the Merge Fraction Button. Smooths the merge fraction for the presently selected dataset.
        try:
            w_range = [float(self.txt_zero_range_min.text()) , float(self.txt_zero_range_max.text())]
            w_idxs = [np.abs(self.TA_matrix_wavelengths - w_range[0]).argmin(), np.abs(self.TA_matrix_wavelengths - w_range[1]).argmin()]
            nm_per_pixel = (np.max(self.TA_matrix_wavelengths) - np.min(self.TA_matrix_wavelengths)) / len(self.TA_matrix_wavelengths) # Determine smoothing window (approx 2nm)
            smoothing_window_size = int(2 / nm_per_pixel)
            if smoothing_window_size < 1: smoothing_window_size = 1   
            for w_idx in w_idxs:                                                                                        # Perform Smoothing
                if (w_idx > 2 * smoothing_window_size) and (w_idx < len(self.TA_matrix_wavelengths) - 2 * smoothing_window_size):   # Avoid edges
                    for _ in range(4):                                                                                  # Repeat smoothing 4 times
                        for ww_idx in range(w_idx - smoothing_window_size, w_idx + smoothing_window_size):
                            start = ww_idx - smoothing_window_size
                            end = ww_idx + smoothing_window_size
                            self.all_probe_counts[self.presently_selected_idx, ww_idx] = np.sum(self.all_probe_counts[self.presently_selected_idx, start:end]) / (2 * smoothing_window_size)
            self.sqrt_all_probe_counts[:] = np.sqrt(self.all_probe_counts)                                              # Recalculate square roots
            sum_sqrt = np.sum(self.sqrt_all_probe_counts, axis=0)
            sum_sqrt[sum_sqrt == 0] = 1e-9
            for i in range(len(self.TA_matrix_input_filenames)):
                self.fraction_in_merge_each_probe_spectrum[i,:] = self.sqrt_all_probe_counts[i,:] / sum_sqrt
            self.update_TA_matrix_plot_via_dropdown()                                                  # Update the plots
            self.status_bar.showMessage("Smoothing applied.")
        except ValueError:
            self.status_bar.showMessage("Error: Invalid input for range.")




    def merge_TA_data(self):                                                                #Merges all datasets based on the calculated fractions and saves to disk.
            self.status_bar.showMessage("Merging data...")
            QApplication.processEvents()
            first_TA_matrix = TA_sh.get_TA_matrix(self.TA_matrix_input_filenames[0])
            merged_TA_data = first_TA_matrix.copy()
            merged_TA_data[1:,1:] = first_TA_matrix[1:,1:] * self.fraction_in_merge_each_probe_spectrum[0,:]    # Initialize with first dataset weighted
            
            final_merged_wavelengths = first_TA_matrix[0,1:]
            final_merged_delay_times = first_TA_matrix[1:,0]
            for i in range(1, len(self.TA_matrix_input_filenames)):                       # Loop through rest
                next_filename = self.TA_matrix_input_filenames[i]
                next_TA_matrix = TA_sh.get_TA_matrix(next_filename)
                next_w = next_TA_matrix[0,1:]
                next_t = next_TA_matrix[1:,0]
                needs_interp = (np.abs(next_w - final_merged_wavelengths) > 0.01).any() or (np.abs(next_t - final_merged_delay_times) > 0.1).any()  # Check if interpolation is needed
                if needs_interp:
                    print(f'Interpolating {next_filename}...')
                    interp_matrix = self.Interpolate_TA_matrix(next_filename, final_merged_wavelengths, final_merged_delay_times)
                    merged_TA_data[1:,1:] += interp_matrix[1:,1:] * self.fraction_in_merge_each_probe_spectrum[i,:]
                else:
                    print(f'Merging {next_filename} (No interpolation needed)...')
                    merged_TA_data[1:,1:] += next_TA_matrix[1:,1:] * self.fraction_in_merge_each_probe_spectrum[i,:]
            try:
                output_filename = self.TA_matrix_input_filenames[0]+'.merged.csv'
                np.savetxt(output_filename, merged_TA_data, delimiter=',')
                self.status_bar.showMessage(f"Merge Complete! Saved to {output_filename}")
                print(f'File saved: {output_filename}')
            except Exception as e:
                self.status_bar.showMessage(f"Save Failed: {e}")


    def Interpolate_TA_matrix(self, TA_matrix_input, new_ws, new_ts):
        TA_matrix = TA_sh.get_TA_matrix(TA_matrix_input)
        old_ws = TA_matrix[0,1:]
        old_ts = TA_matrix[1:,0]
        print('Starting an Interpolation. This will take a minute or two. Be patient.')
        TA_matrix_interpolator = RegularGridInterpolator((old_ts, old_ws), TA_matrix[1:,1:], bounds_error=False, fill_value=None)  # Create the Interpolator: bounds_error=False allows extrapolation or handling points outside the grid without crashing fill_value=None tells it to extrapolate values outside the bounds
        TA_matrix_interpolated = np.zeros([len(new_ts)+1, len(new_ws)+1])
        grid_t, grid_w = np.meshgrid(new_ts, new_ws, indexing='ij') # Create Grids
        TA_matrix_interpolated[0,1:]  = new_ws
        TA_matrix_interpolated[1:,0]  = new_ts
        TA_matrix_interpolated[1:,1:] = TA_matrix_interpolator((grid_t, grid_w ))
        print("Interpolation successful.")
        return TA_matrix_interpolated




class Custom_QT5_Toolbar(NavigationToolbar2QT):
    def __init__(self, canvas, parent_pyqt5_QMainWindow, coordinates=True):
        super().__init__(canvas, parent_pyqt5_QMainWindow, coordinates)
        self.parent_pyqt5_QMainWindow = parent_pyqt5_QMainWindow
    def home(self, *args, **kwargs):
        # Call the Original Home Behavior. This performs the actual view reset in Matplotlib
        super().home(*args, **kwargs)
        # Now I can force my own view
        self.parent_pyqt5_QMainWindow.pcolormesh_transects.TA_image_axis_han.set_xlim(self.parent_pyqt5_QMainWindow.pcolormesh_transects.home_xlim)
        self.parent_pyqt5_QMainWindow.pcolormesh_transects.TA_image_axis_han.set_ylim(self.parent_pyqt5_QMainWindow.pcolormesh_transects.home_ylim)
        self.parent_pyqt5_QMainWindow.canvas.draw()




# For Matt Sfeirs TA beamline the hdf5 files hold the probe counts in 'Spectra/Sweep_0_Probe_Spectrum'
# For the Astrella TA system the probe counts are located in 
def create_merge_ratio_between_2_TA_probe_Matrices(TA_probe1_counts_filename, TA_probe2_counts_filename, final_interpolated_wavelengths, probe1_blackout, probe2_blackout):
    # Obtain the probe counts for the probe1 
    blue_probe_counts = np.flip( TA_sh.load_hdf5_data(TA_probe1_counts_filename,'Spectra/Sweep_0_Probe_Spectrum') )                     #Obtain the blue probe spectrum 
    blue_probe_wavelengths = np.flip( TA_sh.load_hdf5_data(TA_probe1_counts_filename,'Average')[0,1:] )
    spline_interpolator = CubicSpline(blue_probe_wavelengths, blue_probe_counts, extrapolate = True)
    blue_probe_counts_interpolated = spline_interpolator(final_interpolated_wavelengths)

    # Obtain the probe counts for the red spectrum
    red_probe_counts = np.flip( TA_sh.load_hdf5_data(TA_probe2_counts_filename,'Spectra/Sweep_0_Probe_Spectrum') )                     #Obtain the red probe spectrum
    red_probe_wavelengths = np.flip( TA_sh.load_hdf5_data(TA_probe2_counts_filename,'Average')[0,1:] )
    spline_interpolator = CubicSpline(red_probe_wavelengths, red_probe_counts, extrapolate = True)
    red_probe_counts_interpolated = spline_interpolator(final_interpolated_wavelengths)

    # Merge the -DT/T data from the blue and red spectra 
    merge_percentage_red = np.sqrt(red_probe_counts_interpolated) / ( np.sqrt(red_probe_counts_interpolated) + np.sqrt(blue_probe_counts_interpolated) ) 
    merge_percentage_red[final_interpolated_wavelengths > 700] = 1       #This is done to avoid including the enormous error in the blue spectrum above 700 nm






# Define a hook to ensure exceptions are caught by the debugger/terminal
def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    print("Uncaught exception", file=sys.stderr)
    sys.__excepthook__(exc_type, exc_value, exc_traceback)

sys.excepthook = handle_exception



# if __name__ == "__main__":       # This line is asking: "Is this file the one that started the process?"  We can have only one QApplication instance per process, so we need to ensure this is the main file.
#     # Check if a filename was provided in the command line arguments
#     if len(sys.argv) > 1:
#         TA_matrix_input_filenames = sys.argv[1]
#     else:
#         # Fallback default if no file is provided
#         TA_matrix_input_filenames = ['HHHF_Zn_heme_ZnCl_p425nm_blue_300uW.h5.t0_corr.csv','HHHF_Zn_heme_ZnCl_p425nm_red_300uW.h5.t0_corr.csv']  # default input for testing:
#     # Start the application
#     window = TA_merge_matrix_GUI(TA_matrix_input_filenames) #The QApplication instance must be created before this.  The init method recognizes QApplication already exists and uses it.
    



#!/usr/bin/env python                    #this enables a user to run the file by typing only it's name (no need for python prefix)

"""    block comment
Created 20251202

@author: dturney
"""


# TA_data are assumed to be 2D matrices with wavelengths along the 1st row, and with probe delay times along the 1st column

import os
import time
import h5py
import numpy as np
from scipy.ndimage import generic_filter

# Matplotlib Imports (The Plotting Library)
import matplotlib
matplotlib.use('qt5agg')
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

# PyQt5 Imports (The GUI Library)
from PyQt5.QtWidgets import (QGridLayout, QLineEdit, QLabel, QPushButton, QGroupBox)
from PyQt5.QtCore import QTimer, Qt

class embedded_matplotlib_pcolormesh_transects():
    def __init__(self, parent_pyqt5_QMainWindow, frac_horiz=1.0, frac_vert=1.0):
        self.parent_pyqt5_QMainWindow = parent_pyqt5_QMainWindow

        scale_axes = np.array([frac_horiz, frac_vert, frac_horiz, frac_vert])  #This is so the caller can make the pcolormesh and transect axes smaller as needed.

        ### Create Axes (Using your exact coordinates) 
        main_TA_image_axes = np.array([0.06, 1-(1-0.31)*frac_vert, 0.66, 0.685*frac_vert])    #[left   bottom    fraction_horizontal   fraction_vertical ]
        horiz_transect_axes = np.array([0.06, 1-(1-0.05)*frac_vert, 0.66, 0.20*frac_vert])    #[left   bottom    fraction_horizontal   fraction_vertical ]
        vert_transect_axes = np.array([0.79, 1-(1-0.31)*frac_vert, 0.2, 0.685*frac_vert])     #[left   bottom    fraction_horizontal   fraction_vertical ]
        colorbar_axes      = np.array([0.75, 1-(1-0.22)*frac_vert, 0.23, 0.03*frac_vert])     #[left   bottom    fraction_horizontal   fraction_vertical ]

        # Main TA Image
        self.TA_image_axis_han = self.parent_pyqt5_QMainWindow.fig_han.add_axes(main_TA_image_axes )    #[left bottom fraction_horizontal  fraction_vertical ]
        self.TA_image_axis_han.set_ylabel('delay time / ps')
        self.TA_image_axis_han.set_xlabel('wavelength / nm')

        # Horizontal Transect (Bottom)
        self.ax_horiz = self.parent_pyqt5_QMainWindow.fig_han.add_axes(horiz_transect_axes, sharex=self.TA_image_axis_han)  #[left bottom fraction_horizontal  fraction_vertical ]
        self.ax_horiz.grid(True, linestyle=':')
        self.ax_horiz.set_xlabel('wavelength / nm')
        
        # Vertical Transect (Right)
        self.ax_vert = self.parent_pyqt5_QMainWindow.fig_han.add_axes(vert_transect_axes, sharey=self.TA_image_axis_han)  #[left bottom fraction_horizontal  fraction_vertical ]
        self.ax_vert.set_ylabel('delay time / ps', labelpad=2)
        self.ax_vert.grid(True, linestyle=':')
        self.ax_vert.ticklabel_format(axis='x', style='sci', scilimits=(0,0))

        # Colorbar Axis
        self.axis_colorbar = self.parent_pyqt5_QMainWindow.fig_han.add_axes(colorbar_axes)   #[left bottom fraction_horizontal  fraction_vertical ]

        # Initial Plotting of the TA Image
        min_max = [np.percentile(self.parent_pyqt5_QMainWindow.TA_data.flatten(),2), np.percentile(self.parent_pyqt5_QMainWindow.TA_data.flatten(),98)]
        self.TA_colormap = create_TA_Blue_White_Red_Black_colormap(min_max)
        # Plot the TA data as a pcolormesh (like imshow but with correct axis scaling)
        self.pclrmsh = self.TA_image_axis_han.pcolormesh(self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths, self.parent_pyqt5_QMainWindow.TA_matrix_delay_times, self.parent_pyqt5_QMainWindow.TA_data, vmin=min_max[0], vmax=min_max[1], cmap=self.TA_colormap, rasterized=True)
        self.home_xlim = self.TA_image_axis_han.get_xlim()
        self.home_ylim = self.TA_image_axis_han.get_ylim()
        # Colormap and colorbar setup
        self.cbar = self.parent_pyqt5_QMainWindow.fig_han.colorbar(self.pclrmsh, cax=self.axis_colorbar, orientation='horizontal')
        self.cbar.ax.tick_params(axis='x',  rotation=0)
        self.cbar.ax.ticklabel_format(style='sci', axis='x', scilimits=(0.1, 10))
        self.cbar.set_label(r'-$\Delta$T / T')

        # Crosshairs (Initialized at index)
        self.crosshair_v_idx = self.parent_pyqt5_QMainWindow.TA_data_cols // 2
        self.crosshair_h_idx = self.parent_pyqt5_QMainWindow.TA_data_rows - 5
        self.vert_cut_num_ave = 1
        self.horiz_cut_num_ave = 1
        self.crosshair_v = self.TA_image_axis_han.axvline(self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths[self.crosshair_v_idx], color='k', linestyle='--', alpha=0.2)
        self.crosshair_h = self.TA_image_axis_han.axhline(self.parent_pyqt5_QMainWindow.TA_matrix_delay_times[self.crosshair_h_idx], color='k', linestyle='--', alpha=0.2)

        # Transect Lines (Empty/Initial)
        self.line_horiz, = self.ax_horiz.plot(self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths, self.parent_pyqt5_QMainWindow.TA_data[self.crosshair_h_idx,:], color='k', linewidth=1)
        self.red_dot_horiz = self.ax_horiz.scatter(self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths[self.crosshair_v_idx], self.parent_pyqt5_QMainWindow.TA_data[self.crosshair_h_idx, self.crosshair_v_idx], color='red', s=40, zorder=2)
        self.line_vert, = self.ax_vert.plot(self.parent_pyqt5_QMainWindow.TA_data[:,self.crosshair_v_idx], self.parent_pyqt5_QMainWindow.TA_matrix_delay_times, color='k', linewidth=1)
        self.red_dot_vert = self.ax_vert.scatter(self.parent_pyqt5_QMainWindow.TA_data[self.crosshair_h_idx, self.crosshair_v_idx], self.parent_pyqt5_QMainWindow.TA_matrix_delay_times[self.crosshair_h_idx], color='red', s=40, zorder=2)

        # create pyqt5 control 1: Axis Limits 
        self.linlog_threshold_time_axis = 0.01  # ps
        grp_ax_limits = QGroupBox("TA Plot Axis Limits")
        grp_ax_limits.setStyleSheet('QLabel { font-size: 11pt; } QLineEdit { font-size: 11pt; }')
        grid_ax_limits = QGridLayout()
        grid_ax_limits.setVerticalSpacing(2)
        grid_ax_limits.setContentsMargins(2, 5, 2, 2)                   # row,   column wthin row,   # of rows to occupy,    # of columns to occupy)
        # Wavelength Axis Limits
        grid_ax_limits.addWidget(QLabel("Wavelength Max:"), 0, 0)       # row ,  column)
        self.txt_w_max = QLineEdit(f"{np.max(self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths):4.2f}")
        self.txt_w_max.returnPressed.connect(self.update_axis_limits)
        grid_ax_limits.addWidget(self.txt_w_max, 0, 1)
        grid_ax_limits.addWidget(QLabel("Wavelength Min:"), 1, 0)
        self.txt_w_min = QLineEdit(f"{np.min(self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths):4.2f}")
        self.txt_w_min.returnPressed.connect(self.update_axis_limits)
        grid_ax_limits.addWidget(self.txt_w_min, 1, 1)
        # Time Axis Limits
        grid_ax_limits.addWidget(QLabel("Time Max:"), 2, 0)
        self.txt_t_max = QLineEdit(f"{np.max(self.parent_pyqt5_QMainWindow.TA_matrix_delay_times):5.2f}")
        self.txt_t_max.returnPressed.connect(self.update_axis_limits)
        grid_ax_limits.addWidget(self.txt_t_max, 2, 1)
        grid_ax_limits.addWidget(QLabel("Time Min:"), 3, 0)
        self.txt_t_min = QLineEdit(f"{np.min(self.parent_pyqt5_QMainWindow.TA_matrix_delay_times):5.2f}")
        self.txt_t_min.returnPressed.connect(self.update_axis_limits)
        grid_ax_limits.addWidget(self.txt_t_min, 3, 1)
        grp_ax_limits.setLayout(grid_ax_limits)
        self.parent_pyqt5_QMainWindow.control_panel_layout.addWidget(grp_ax_limits)


        # create pyqt5 control 2: Color Scale
        grp_colormap = QGroupBox("Color Map Scaling")
        grp_colormap.setStyleSheet('QLabel { font-size: 11pt; } QLineEdit { font-size: 11pt; }')
        grid_clrmap_limits = QGridLayout()   
        grid_clrmap_limits.setVerticalSpacing(2)
        grid_clrmap_limits.setContentsMargins(2, 5, 2, 2)                      # row,   column wthin row,   # of rows to occupy,    # of columns to occupy)
        grid_clrmap_limits.addWidget(QLabel("Colormap max:"), 0, 0)            # row ,  column)
        clim = self.pclrmsh.get_clim()
        self.txt_cmap_max = QLineEdit( f"{clim[1]:.5f}" )
        self.txt_cmap_max.returnPressed.connect(self.update_colormap_manual)
        grid_clrmap_limits.addWidget(self.txt_cmap_max, 0, 1)
        grid_clrmap_limits.addWidget(QLabel("Colormap min:"), 1, 0)
        clim = self.pclrmsh.get_clim()
        self.txt_cmap_min = QLineEdit( f"{clim[0]:.5f}" )
        self.txt_cmap_min.returnPressed.connect(self.update_colormap_manual)
        grid_clrmap_limits.addWidget(self.txt_cmap_min, 1, 1)     
        grp_colormap.setLayout(grid_clrmap_limits)
        self.parent_pyqt5_QMainWindow.control_panel_layout.addWidget(grp_colormap)


        # create pyqt5 control 3: Time Axis Lin-Log Scaling
        self.linear_threshold_time_axis = 0.01
        grp_scale = QGroupBox("Time Axis Scaling")
        grp_scale.setStyleSheet('QLabel { font-size: 11pt; } QLineEdit { font-size: 11pt; } QPushButton {font-size: 11pt; }')
        grid_linlog = QGridLayout()
        grid_linlog.setVerticalSpacing(2)
        grid_linlog.setContentsMargins(2, 5, 2, 2)                      # row,   column wthin row,   # of rows to occupy,    # of columns to occupy)
        grid_linlog.addWidget(QLabel("Lin-Log limit (ps):"), 0, 0)
        self.txt_linlog_thresh = QLineEdit(str(self.linlog_threshold_time_axis))
        self.txt_linlog_thresh.returnPressed.connect(self.update_linlog_thresh)
        grid_linlog.addWidget(self.txt_linlog_thresh, 0, 1)
        self.btn_linlog_time = QPushButton("Lin-Log y-axis")
        self.btn_linlog_time.setCheckable(True)
        self.btn_linlog_time.setStyleSheet("QPushButton { background-color: #F8F8F8; border: 1px solid #ADADAD; border-radius: 4px; padding: 3px;}  QPushButton:checked {font-size: 11pt; background-color: #E0E0E0; border: 1px solid #ADADAD;}")
        self.btn_linlog_time.clicked.connect(self.on_linlog_toggle)
        grid_linlog.addWidget(self.btn_linlog_time, 1, 0, 1, 2)         # row,   column wthin row,   # of rows to occupy,    # of columns to occupy)
        grp_scale.setLayout(grid_linlog)
        self.parent_pyqt5_QMainWindow.control_panel_layout.addWidget(grp_scale)


        # create pyqt5 control 4: Averaging
        grp_avg = QGroupBox("Transect Averaging")
        grp_avg.setStyleSheet('QLabel { font-size: 11pt; } QLineEdit { font-size: 11pt; }')
        grid_avg = QGridLayout()
        grid_avg.setVerticalSpacing(2)
        grid_avg.setContentsMargins(2, 5, 2, 2)
        grid_avg.addWidget(QLabel("Horiz Avg (rows):"), 0, 0)
        self.txt_h_avg = QLineEdit("1")
        self.txt_h_avg.returnPressed.connect(self.update_transects)
        grid_avg.addWidget(self.txt_h_avg, 0, 1)
        grid_avg.addWidget(QLabel("Vert Avg (cols):"), 1, 0)
        self.txt_v_avg = QLineEdit("1")
        self.txt_v_avg.returnPressed.connect(self.update_transects)
        grid_avg.addWidget(self.txt_v_avg, 1, 1)
        grp_avg.setLayout(grid_avg)
        self.parent_pyqt5_QMainWindow.control_panel_layout.addWidget(grp_avg)

        # Set custom formatter for the toolbar status message
        self.TA_image_axis_han.format_coord = self.format_toolbar_status_msg_TA_matrix
        self.ax_horiz.format_coord = self.format_toolbar_status_msg_horiz_transect
        self.ax_vert.format_coord = self.format_toolbar_status_msg_vert_transect

        # # Setup the QTimer
        self.timer = QTimer()
        self.timer.timeout.connect(self.run_every_500ms) # Connect the 'timeout' signal to your function
        self.timer.start(500)  # Start the timer (Interval in milliseconds)


    #########################
    ####### Update every 500 ms
    def run_every_500ms(self):
        #This function gets called automatically every 500ms.
        xlim = self.TA_image_axis_han.get_xlim()
        ylim = self.TA_image_axis_han.get_ylim()
        try:
            if np.abs(xlim[0] - float(self.txt_w_min.text())) > 0.1 or np.abs(xlim[1] - float(self.txt_w_max.text())) > 0.1 or np.abs(ylim[0] - float(self.txt_t_min.text())) > 0.1 or np.abs(ylim[1] - float(self.txt_t_max.text())) > 0.1:
                if self.parent_pyqt5_QMainWindow.focusWidget() != self.txt_w_min and self.parent_pyqt5_QMainWindow.focusWidget() != self.txt_w_max and self.parent_pyqt5_QMainWindow.focusWidget() != self.txt_t_min and self.parent_pyqt5_QMainWindow.focusWidget() != self.txt_t_max:
                    self.txt_w_min.setText(f"{xlim[0]:.2f}")
                    self.txt_w_max.setText(f"{xlim[1]:.2f}")
                    self.txt_t_min.setText(f"{ylim[0]:.2f}")
                    self.txt_t_max.setText(f"{ylim[1]:.2f}")
                    self.update_transects()
        except ValueError:
            pass



    ##############################################
    ####### Callbacks for Interactivity 

    def update_transects(self):
        ### Updates the side plots based on current crosshair positions and averaging settings.        
        try:
            self.horiz_cut_num_ave = int(self.txt_h_avg.text())
            self.vert_cut_num_ave = int(self.txt_v_avg.text())
        except ValueError:
            self.parent_pyqt5_QMainWindow.status_bar.showMessage("Invalid Transect Input")
        # Determine Padding for Averaging, with safe bounds checking
        pad_h = 0
        if self.horiz_cut_num_ave > 1:
            pad_h = int((self.horiz_cut_num_ave - 1) // 2) # simplified logic
        pad_v = 0
        if self.vert_cut_num_ave > 1:
            pad_v = int((self.vert_cut_num_ave - 1) // 2)
        # Safe bounds check
        h_start = max(0, self.crosshair_h_idx - pad_h)
        h_end = min(self.parent_pyqt5_QMainWindow.TA_data_rows, self.crosshair_h_idx + pad_h + 1)
        v_start = max(0, self.crosshair_v_idx - pad_v)
        v_end = min(self.parent_pyqt5_QMainWindow.TA_data_cols, self.crosshair_v_idx + pad_v + 1)

        # Find current zoom xlim and ylim so you can automatically rescale transects
        xlim = self.TA_image_axis_han.get_xlim()
        w_mask = (self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths >= xlim[0]) & (self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths <= xlim[1])
        ylim = self.TA_image_axis_han.get_ylim()
        t_mask = (self.parent_pyqt5_QMainWindow.TA_matrix_delay_times >= ylim[0]) & (self.parent_pyqt5_QMainWindow.TA_matrix_delay_times <= ylim[1])

        # Horizontal Transect (Slicing the matrix)
        # Average the rows around the crosshair
        averaged_spectrum = np.mean(self.parent_pyqt5_QMainWindow.TA_data[h_start:h_end, :], axis=0)
        self.line_horiz.set_ydata(averaged_spectrum)
        # Rescale view
        self.ax_horiz.set_ylim([averaged_spectrum[w_mask].min(), averaged_spectrum[w_mask].max()])
        # Update red dot
        self.red_dot_horiz.set_offsets([self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths[self.crosshair_v_idx], averaged_spectrum[self.crosshair_v_idx]])

        # Vertical Transect
        # Average the columns around the crosshair
        averaged_transient = np.mean(self.parent_pyqt5_QMainWindow.TA_data[:, v_start:v_end], axis=1)
        self.line_vert.set_xdata(averaged_transient)
        # Rescale view
        self.ax_vert.set_xlim([averaged_transient[t_mask].min(), averaged_transient[t_mask].max()])
        # Update red dot
        self.red_dot_vert.set_offsets([averaged_transient[self.crosshair_h_idx], self.parent_pyqt5_QMainWindow.TA_matrix_delay_times[self.crosshair_h_idx]])

        self.parent_pyqt5_QMainWindow.canvas.draw()


    def update_crosshair_lines(self):
        ### Updates just the dashed lines on the heatmap.
        # Get actual float values
        w_val = self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths[self.crosshair_v_idx]
        t_val = self.parent_pyqt5_QMainWindow.TA_matrix_delay_times[self.crosshair_h_idx]
        self.crosshair_v.set_xdata([w_val, w_val])
        self.crosshair_h.set_ydata([t_val, t_val])
        self.parent_pyqt5_QMainWindow.canvas.draw()
    

    def update_axis_limits(self):
        try:
            self.allow_TA_matrix_axis_limit_update = False
            self.TA_image_axis_han.set_xlim(float(self.txt_w_min.text()), float(self.txt_w_max.text()))
            self.TA_image_axis_han.set_ylim(float(self.txt_t_min.text()), float(self.txt_t_max.text()))
            self.parent_pyqt5_QMainWindow.canvas.draw()
            if self.parent_pyqt5_QMainWindow.status_bar.currentMessage()=="Invalid Axis Limit Input":
                self.parent_pyqt5_QMainWindow.status_bar.showMessage("")
        except ValueError:
            self.allow_TA_matrix_axis_limit_update = True
            self.parent_pyqt5_QMainWindow.status_bar.showMessage("Invalid Axis Limit Input")


    def update_colormap_manual(self):
        try:
            self.TA_colormap = create_TA_Blue_White_Red_Black_colormap([float(self.txt_cmap_min.text()), float(self.txt_cmap_max.text())])
            self.pclrmsh.set_cmap(self.TA_colormap)
            self.pclrmsh.set_clim(float(self.txt_cmap_min.text()), float(self.txt_cmap_max.text()))
            self.parent_pyqt5_QMainWindow.canvas.draw()
            if self.parent_pyqt5_QMainWindow.status_bar.currentMessage()=="Invalid Colormap Input":
                self.parent_pyqt5_QMainWindow.status_bar.showMessage("")
        except:
            self.parent_pyqt5_QMainWindow.status_bar.showMessage("Invalid Colormap Input")
        

    def on_linlog_toggle(self, checked):
        try:
            self.linlog_threshold_time_axis = float(self.txt_linlog_thresh.text())
        except ValueError:
            self.parent_pyqt5_QMainWindow.status_bar.showMessage("Invalid LinLog Threshold Input")
            self.btn_linlog_time.setChecked(False)
            return
        self.parent_pyqt5_QMainWindow.status_bar.showMessage("")
        if checked:
            self.TA_image_axis_han.set_yscale('symlog', linthresh=self.linlog_threshold_time_axis)
        else:
            self.TA_image_axis_han.set_yscale('linear')
        self.parent_pyqt5_QMainWindow.canvas.draw()



    def update_linlog_thresh(self):
        try:
            self.allow_TA_matrix_axis_limit_update = False
            self.linlog_threshold_time_axis = float(self.txt_linlog_thresh.text())
            # If currently in linlog mode, update immediately
            if self.btn_linlog_time.isChecked():
                self.TA_image_axis_han.set_yscale('symlog', linthresh=self.linlog_threshold_time_axis)
                self.parent_pyqt5_QMainWindow.canvas.draw()
            if self.parent_pyqt5_QMainWindow.status_bar.currentMessage()=="Invalid LinLog Threshold Input":
                self.parent_pyqt5_QMainWindow.status_bar.showMessage("")
            self.allow_TA_matrix_axis_limit_update = True
        except ValueError:
            self.allow_TA_matrix_axis_limit_update = True
            self.parent_pyqt5_QMainWindow.status_bar.showMessage("Invalid LinLog Threshold Input")



    def format_toolbar_status_msg_TA_matrix(self, w_mouse, t_mouse):
        # Custom formatter for the toolbar message. Matplotlib calls this automatically with x, y coords.
        try:
            z_mouse = self.parent_pyqt5_QMainWindow.TA_matrix_interpolator([t_mouse, w_mouse])[0]    # Look up Z-value
            w_mouse_str = f"{w_mouse:>6.1f}"    # convert to string
            t_mouse_str = f"{t_mouse:>6.1f}"    # convert to string
            z_mouse_str = f"{z_mouse:>+3.8f}"   # convert to string
            w_crosshair = self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths[self.crosshair_v_idx]
            t_crosshair = self.parent_pyqt5_QMainWindow.TA_matrix_delay_times[self.crosshair_h_idx]
            z_val_crosshairs = self.parent_pyqt5_QMainWindow.TA_matrix_interpolator([t_crosshair, w_crosshair])[0]
            w_crosshair_str = f"{w_crosshair:>5.1f}"
            t_crosshair_str = f"{t_crosshair:>6.1f}"
            z_crosshair_str = f"{z_val_crosshairs:>+3.8f}"
        except Exception:
            return ""
        msg = f"Mouse: w={w_mouse_str} nm t={t_mouse_str} ps z={z_mouse_str}\nCrosshairs: w={w_crosshair_str}nm, t={t_crosshair_str}ps, z={z_crosshair_str}"
        return msg

    def format_toolbar_status_msg_horiz_transect(self, w_mouse, z_mouse):
        # Custom formatter for the toolbar message. Matplotlib calls this automatically with x, y coords.
        try:
            t_mouse = self.parent_pyqt5_QMainWindow.TA_matrix_delay_times[self.crosshair_h_idx]
            w_mouse_str = f"{w_mouse:>6.1f}"    # convert to string
            t_mouse_str = f"{t_mouse:>6.1f}"    # convert to string
            z_mouse_on_transect = self.parent_pyqt5_QMainWindow.TA_matrix_interpolator([t_mouse, w_mouse])[0]
            z_mouse_str = f"{z_mouse_on_transect:>+3.8f}"   # convert to string
            w_crosshair = self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths[self.crosshair_v_idx]
            t_crosshair = self.parent_pyqt5_QMainWindow.TA_matrix_delay_times[self.crosshair_h_idx]
            z_val_crosshairs = self.parent_pyqt5_QMainWindow.TA_matrix_interpolator([t_crosshair, w_crosshair])[0]
            w_crosshair_str = f"{w_crosshair:>5.1f}"
            t_crosshair_str = f"{t_crosshair:>6.1f}"
            z_crosshair_str = f"{z_val_crosshairs:>+3.8f}"
        except Exception:
            return ""
        msg = f"Mouse: w={w_mouse_str} nm t={t_mouse_str} ps z={z_mouse_str}\nCrosshairs: w={w_crosshair_str}nm, t={t_crosshair_str}ps, z={z_crosshair_str}"
        return msg
    
    def format_toolbar_status_msg_vert_transect(self, z_mouse, t_mouse):
        # Custom formatter for the toolbar message. Matplotlib calls this automatically with x, y coords.
        try:
            w_mouse = self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths[self.crosshair_v_idx]
            w_mouse_str = f"{w_mouse:>6.1f}"    # convert to string
            t_mouse_str = f"{t_mouse:>6.1f}"    # convert to string
            z_mouse_on_transect = self.parent_pyqt5_QMainWindow.TA_matrix_interpolator([t_mouse, w_mouse])[0]
            z_mouse_str = f"{z_mouse_on_transect:>+3.8f}"   # convert to string
            w_crosshair = self.parent_pyqt5_QMainWindow.TA_matrix_wavelengths[self.crosshair_v_idx]
            t_crosshair = self.parent_pyqt5_QMainWindow.TA_matrix_delay_times[self.crosshair_h_idx]
            z_val_crosshairs = self.parent_pyqt5_QMainWindow.TA_matrix_interpolator([t_crosshair, w_crosshair])[0]
            w_crosshair_str = f"{w_crosshair:>5.1f}"
            t_crosshair_str = f"{t_crosshair:>6.1f}"
            z_crosshair_str = f"{z_val_crosshairs:>+3.8f}"
        except Exception:
            return ""
        msg = f"Mouse: w={w_mouse_str} nm t={t_mouse_str} ps z={z_mouse_str}\nCrosshairs: w={w_crosshair_str}nm, t={t_crosshair_str}ps, z={z_crosshair_str}"
        return msg
























def plot_fractional_change_TA_probe_counts(TA_filename):
    sweep_0_probe_counts = load_hdf5_data(TA_filename,'Spectra/Sweep_0_Probe_Spectrum')
    sweep_1_probe_counts = load_hdf5_data(TA_filename,'Spectra/Sweep_1_Probe_Spectrum')
    wavelengths = get_TA_matrix(TA_filename)[0,1:]
    fractional_change = (sweep_1_probe_counts - sweep_0_probe_counts) / sweep_0_probe_counts
    fig_han = plt.figure()
    ax=fig_han.add_axes([0.135,0.15,0.75,0.75])
    ax.plot(wavelengths, fractional_change, 'b', label='fractional change \n probe counts')
    ax2 = ax.twinx()
    ax2.plot(wavelengths, sweep_0_probe_counts, 'g', label = 'probe counts')
    ax.set_xlim(450,800)
    ax.set_ylim(-0.1,0.1)
    ax2.set_ylim(0,3000)
    ax.set_xlabel('wavelength / nm')
    ax.set_ylabel('fractional change', color = 'b')
    ax.tick_params(axis='y', colors='b')
    ax2.set_ylabel('probe counts', color='g')
    ax2.tick_params(axis='y', colors='g')
    ax.set_title(TA_filename)
    plt.show()
    





def TA_matrix_window_average(TA_matrix,window_size):
    TA_matrix = get_TA_matrix(TA_matrix)
    TA_matrix_window_averaged = TA_matrix.copy()
    TA_matrix_window_averaged[1:,1:] = generic_filter(TA_matrix[1:,1:], np.mean, size=window_size, mode='constant', cval=0)
    return TA_matrix_window_averaged



def create_TA_Blue_White_Red_colormap(min_max):
    #white_fractional_location is a number between 0 and 1
    TA_colormap = np.ones([100,4])
    white_fractional_location = (0.0 - min_max[0])/ (min_max[1] - min_max[0])
    index_white = int(np.round(white_fractional_location*100))

    # setup the blue portion of the colormap
    for i in range(0, index_white):
        TA_colormap[i,0] = i/index_white
        TA_colormap[i,1] = i/index_white

    # setup the red portion of the colormap
    for i in range(index_white, 100):
        TA_colormap[i,1] = (100-i)/(100-index_white)
        TA_colormap[i,2] = (100-i)/(100-index_white)

    colorlist = TA_colormap.tolist()
    cmap_name = "TA_blue_white_red_colormap"
    TA_colormap = LinearSegmentedColormap.from_list(cmap_name, colorlist, N=120)
    return TA_colormap




def create_TA_Blue_White_Red_Black_colormap(min_max):
    if isinstance(min_max[0], str):
        min = float(min_max[0])
        max = float(min_max[1])
    else:
        min = min_max[0]
        max = min_max[1]
    #white_fractional_location is a number between 0 and 1
    TA_colormap = np.ones([100,4])
    white_fractional_location = (0.0 - min)/ (max - min)
    index_white = int(np.round(white_fractional_location*100))

    # setup the blue portion of the colormap
    for i in range(0, index_white):
        TA_colormap[i,0] = i/index_white
        TA_colormap[i,1] = i/index_white

    # setup the red portion of the colormap
    full_red_index = int((100 - index_white)*2/4 + index_white)
    for i in range(index_white, full_red_index):
        TA_colormap[i,1] = (full_red_index-i)/(full_red_index-index_white)
        TA_colormap[i,2] = (full_red_index-i)/(full_red_index-index_white)
    for i in range(full_red_index, 100):
        TA_colormap[i,1] = 0.0
        TA_colormap[i,2] = 0.0

    # setup the red-to-black portion of the colormap
    for i in range(full_red_index, 100):
        TA_colormap[i,0] = (100-i)*3/4 / (100-full_red_index) + 0.25

    colorlist = TA_colormap.tolist()
    cmap_name = "TA_blue_white_red_colormap"
    TA_colormap = LinearSegmentedColormap.from_list(cmap_name, colorlist, N=120)
    return TA_colormap




def get_TA_matrix(TA_matrix):
    ## Handle the input if it's a raw matrix or if it's a filename that needs to be passed to a function to load a raw matrix
    if isinstance(TA_matrix, np.ndarray):
        return TA_matrix
    if  np.array([TA_matrix[-5:] == '.hdf5' , TA_matrix[-5:] == '.HDF5',  TA_matrix[-3:] == '.h5',  TA_matrix[-3:] == '.H5' ]).any() :
        TA_matrix = load_hdf5_data(TA_matrix, 'Average')
    elif TA_matrix[-4:] == '.csv':
        TA_matrix = np.loadtxt(TA_matrix, delimiter=',',  ndmin=2)
    return TA_matrix







def get_TA_probe_counts(TA_matrix):
    ## Handle the input if it's a raw matrix or if it's a filename that needs to be passed to a function to load a raw matrix
    if  np.array([TA_matrix[-5:] == '.hdf5' , TA_matrix[-5:] == '.HDF5',  TA_matrix[-3:] == '.h5',  TA_matrix[-3:] == '.H5' ]).any() :
        TA_matrix = load_hdf5_data(TA_matrix, 'Spectra/Sweep_0_Probe_Spectrum')
    elif TA_matrix.find('.hdf5') != -1:
        TA_matrix = load_hdf5_data( TA_matrix[0:TA_matrix.find('.hdf5') + 5], 'Spectra/Sweep_0_Probe_Spectrum')
    elif TA_matrix.find('.HDF5') != -1:
        TA_matrix = load_hdf5_data( TA_matrix[0:TA_matrix.find('.HDF5') + 5], 'Spectra/Sweep_0_Probe_Spectrum')
    elif TA_matrix.find('.h5') != -1:
        TA_matrix = load_hdf5_data( TA_matrix[0:TA_matrix.find('.h5') + 3], 'Spectra/Sweep_0_Probe_Spectrum')
    elif TA_matrix.find('.H5') != -1:
        TA_matrix = load_hdf5_data( TA_matrix[0:TA_matrix.find('.H5') + 3], 'Spectra/Sweep_0_Probe_Spectrum')
    else:
        print('Didnt recognize file format.')    
    
    return TA_matrix






def load_hdf5_data(filename,dataset_path_string):
    
    # the dataset_path_string uses / to delimit the groups and subgroups and datasets, e.g. 'experiment_1/readings/voltage' 
    with h5py.File(filename, 'r') as f:
        # 1. Access the dataset object (this is a pointer to the file, not the data yet)
        dset = f[dataset_path_string][:]
    
    # 2. Convert to NumPy array
    # The [:] slice syntax tells h5py to read the whole dataset into memory
    return dset






def list_hdf5_contents(filename):
    # Opens an HDF5 file and prints its hierarchy (Groups/Datasets) and any attributes (variables) associated with them.

    if not os.path.exists(filename):
        print(f"Error: The file '{filename}' was not found in the current directory.")
        return

    def print_attrs(name, obj):
        # Helper to print attributes (variables) of an object
        if obj.attrs:
            print(f"{'  ' * (indent + 1)}Attributes:")
            for key, val in obj.attrs.items():
                print(f"{'  ' * (indent + 2)}- {key}: {val}")

    def recurse(name, node, indent=0):
        # Determine if it's a Group or Dataset
        if isinstance(node, h5py.Group):
            print(f"{'  ' * indent}📂 Group: {name}")
            
            # Print attributes of the group
            if node.attrs:
                print(f"{'  ' * (indent + 1)}Attributes:")
                for key, val in node.attrs.items():
                    print(f"{'  ' * (indent + 2)}- {key}: {val}")
            
            # Recurse into children
            for child_name, child_node in node.items():
                recurse(child_name, child_node, indent + 1)
                
        elif isinstance(node, h5py.Dataset):
            # Print dataset info (shape and dtype are usually helpful context)
            print(f"{'  ' * indent}📄 Dataset: {name} (Shape: {node.shape}, Type: {node.dtype})")
            
            # Print attributes of the dataset
            if node.attrs:
                print(f"{'  ' * (indent + 1)}Attributes:")
                for key, val in node.attrs.items():
                    print(f"{'  ' * (indent + 2)}- {key}: {val}")

    try:
        with h5py.File(filename, 'r') as f:
            print(f"--- Structure of {filename} ---")
            # Start recursion from the root group
            # We manually iterate root keys to keep the indent logic clean
            if f.attrs:
                print("Root Attributes:")
                for key, val in f.attrs.items():
                    print(f"  - {key}: {val}")
            indent = 0        
            for key, val in f.items():
                recurse(key, val, indent)
                
    except OSError:
        print(f"Error: Could not open '{filename}'. Is it a valid HDF5 file?")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")




"""
Top-level package for github-repo-name.
"""

# 1. Version Handling
# We try to import the version generated by setuptools_scm.
# If the package isn't installed/built yet, we fallback to a default.
try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

# 2. Namespace Flattening
# Import main classes/functions from your source scripts so they are available
# directly at the package level.
# Replace 'MyMainClass' and 'my_utility_function' with the ACTUAL names
# inside your source_scripts.py file.
from .source_scripts import (
    TA_merge_matrix_GUI,
    TA_plot_matrix_GUI,
    TA_t0_correction_and_background_removal_GUI
)

# 3. Define the public API
# This controls what happens if someone does 'from github_repo_name import *'
__all__ = [
    "TA_merge_matrix_GUI",
    "TA_plot_matrix_GUI",
    "TA_t0_correction_and_background_removal_GUI"
]


#!/usr/bin/env python                    #this enables a user to run the file by typing only it's name (no need for python prefix)

"""    block comment
Created 20251202

@author: dturney
"""


# TA_data are assumed to be 2D matrices with wavelengths along the 1st row, and with probe delay times along the 1st column

import sys
import numpy as np
from scipy.interpolate import RegularGridInterpolator

# PyQt5 Imports (The GUI Library)
from PyQt5.QtWidgets import (QApplication, QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QGridLayout, QLineEdit, QLabel, QPushButton, QStatusBar, QGroupBox)
from PyQt5.QtCore import QTimer, Qt

# Matplotlib Imports (The Plotting Library)
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT 
from matplotlib.figure import Figure

# Import the shared functions and classes
import shared_functions_classes as TA_sh


class TA_plot_matrix_GUI():   
    def __init__(self, TA_matrix_input):
        self.app = QApplication(sys.argv)                                           # Create the QApplication instance           
        self.TA_matrix_input = TA_matrix_input
        self.main_pyqt5_window = TA_plot_matrix_main_window(self.TA_matrix_input)   # Create the main pyqt5 window
        self.main_pyqt5_window.show()                                               # Display the window
        sys.exit(self.app.exec())                                                   # Start the event loop        



class TA_plot_matrix_main_window(QMainWindow):    #This nomenclature causes this class to inherit all the methods and properties of the QMainWindow class from PyQT5.
    def __init__(self, TA_matrix_input):
        
        super().__init__() #This executes a lot of code that builds the PyQT5 window using the QMainWindow class.
        
        # pyqt5 window setup: needs the QMainWindow object to be inherited into this class. 
        self.setWindowTitle("TA Matrix Viewer")
        self.dpi = QApplication.primaryScreen().logicalDotsPerInch() #Get screen DPI
        self.window_width_inches = 15
        self.window_height_inches = 10.5
        self.resize(int(self.window_width_inches * self.dpi), int(self.window_height_inches * self.dpi))  # Width, Height in pixels (converted from inches)

        # Create the main pyqt5 object (Qwidget) and then create the geometric layout that will hold a LHS box (for the matplotlib stuff) and RHS box (for the pyqt5 Controls).
        self.main_window_widget = QWidget()
        self.setCentralWidget(self.main_window_widget)
        self.horiz_pyqt5_layout = QHBoxLayout(self.main_window_widget) # Horizontal: Plot on Left, Controls on Right

        # Add the vertical pyqt5 layout to the horizontal (main) layout. Create the geometric layout of the LHS box of the QHBoxLayout window: to hold the matplotlib toolbar on top and the matplotlib figure below. This ensures the toolbar sits right on top of the plot
        self.vertical_layout_widget = QWidget()
        self.vert_pyqt5_layout = QVBoxLayout(self.vertical_layout_widget)   # Vertical pyqt5 layout: toolbar on top, canvas below
        self.vert_pyqt5_layout.setContentsMargins(0, 0, 0, 0) # Remove extra spacing
        self.horiz_pyqt5_layout.addWidget(self.vertical_layout_widget, stretch=4)

        # Create the Control Panel (aka the pyqt5 layout for inserting buttons and textboxes) 
        self.control_panel = QWidget()
        self.control_panel_layout = QVBoxLayout(self.control_panel)
        self.control_panel_layout.setAlignment(Qt.AlignTop)
        self.horiz_pyqt5_layout.addWidget(self.control_panel, stretch=1) # Add the control panel to main horizontal layout (since it's added to the main horizontal layout 1st, it's located on the right side).

        # Create matplotlib Figure and PyQT5 Canvas (actually, it's a pyqt5 imitation of a matplotlib toolbar). 
        self.fig_han = Figure(dpi=self.dpi)   # We don't need to specify figsize here since the canvas will auto-scale to fill the pyqt5 layout space.
        self.canvas = FigureCanvasQTAgg(self.fig_han)
        self.toolbar = Custom_QT5_Toolbar(self.canvas, self)
        self.toolbar.setStyleSheet("QLabel { font-family: 'Courier New', Consolas, monospace; font-size: 13pt; white-space: pre; }") # We have to use specific fonts to get the toolbar message to not wiggle with different number font widths.
        self.vert_pyqt5_layout.addWidget(self.toolbar)
        self.vert_pyqt5_layout.addWidget(self.canvas, stretch=4) # Add Canvas to the main horizontal layout (Takes up most of the space. Since it's added to the main horizontal laout 2nd, it's on the left side).

        # Load the TA matrix data
        self.TA_matrix_input = TA_matrix_input
        self.TA_matrix = TA_sh.get_TA_matrix(self.TA_matrix_input)
        # Extract the wavelengths and time delays
        self.TA_matrix_wavelengths = self.TA_matrix[0,1:]
        self.TA_matrix_delay_times = self.TA_matrix[1:,0]
        # Crop the TA image down to remove the wavelengths and delay times
        self.TA_data = self.TA_matrix[1:,1:]
        self.TA_data_rows, self.TA_data_cols = self.TA_data.shape
        # Create interpolator for the status bar readout
        self.TA_matrix_interpolator = RegularGridInterpolator((self.TA_matrix_delay_times, np.flip(self.TA_matrix_wavelengths)), np.fliplr(self.TA_data), bounds_error=False, fill_value=None)


        # Create the Matplotlib Figure and Canvas
        self.pcolormesh_transects = TA_sh.embedded_matplotlib_pcolormesh_transects(parent_pyqt5_QMainWindow=self, frac_horiz=1.0, frac_vert=1.0)

        # Add simple instructions
        self.fig_han.text(0.76, 0.068, 'Instructions:')
        self.fig_han.text(0.76, 0.03, 'Double mouse click (left button)\nto move crosshairs.')

        # create a pyqt5 button control: Save Transient
        self.btn_save = QPushButton("Save Plotted Transient (.csv)")

        self.btn_save.clicked.connect(self.save_TA_transient)
        self.control_panel_layout.addWidget(self.btn_save)
        self.btn_save.setStyleSheet('QPushButton { font-size: 11pt; }')
        # Set focus so keyboard works immediately
        self.canvas.setFocusPolicy(Qt.ClickFocus)
        self.canvas.setFocus()

        # Initial Draw
        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)

        # Connect Matplotlib Events
        self.canvas.mpl_connect('button_press_event', self.on_TA_matrix_mouse_click)
        self.canvas.mpl_connect('key_press_event', self.on_TA_matrix_up_down_left_right)

    ############################################
    ####### Callbacks for Interactivity with the PyQt5 Control Buttons and Textboxes

    def save_TA_transient(self):
        w_val = self.TA_matrix_wavelengths[self.pcolormesh_transects.crosshair_v_idx]
        if isinstance(self.TA_matrix_input, str):
            filename = f"{self.TA_matrix_input}.transient{w_val:.1f}nm.csv"
        else:
            filename = f"transient{w_val:.1f}nm.csv"
        data = np.column_stack((self.pcolormesh_transects.line_vert.get_xdata(), self.pcolormesh_transects.line_vert.get_ydata()))
        np.savetxt(filename, data, delimiter=',')
        self.status_bar.showMessage(f"Saved: {filename}")


    def on_TA_matrix_up_down_left_right(self, event):
        ### Handles arrow keys.
        # Check boundaries
        if event.key == 'up' and self.pcolormesh_transects.crosshair_h_idx < self.TA_data_rows - 1:
            self.pcolormesh_transects.crosshair_h_idx += 1
        elif event.key == 'down' and self.pcolormesh_transects.crosshair_h_idx > 0:
            self.pcolormesh_transects.crosshair_h_idx -= 1
        elif event.key == 'left' and self.pcolormesh_transects.crosshair_v_idx < self.TA_data_cols - 1:
            self.pcolormesh_transects.crosshair_v_idx += 1
        elif event.key == 'right' and self.pcolormesh_transects.crosshair_v_idx > 0:
            self.pcolormesh_transects.crosshair_v_idx -= 1
        else:
            return # Ignore other keys
        self.pcolormesh_transects.update_crosshair_lines()
        self.pcolormesh_transects.update_transects()


    def on_TA_matrix_mouse_click(self, event):
        ### Handles mouse clicks to move crosshairs.
        if event.inaxes != self.pcolormesh_transects.TA_image_axis_han: return
        # We only care about Double Clicks (per your original script)
        if event.dblclick:
            # Update indices
            self.pcolormesh_transects.crosshair_v_idx = int(np.abs(self.TA_matrix_wavelengths - event.xdata).argmin())
            self.pcolormesh_transects.crosshair_h_idx = int(np.abs(self.TA_matrix_delay_times - event.ydata).argmin())
            self.pcolormesh_transects.update_crosshair_lines()
            self.pcolormesh_transects.update_transects()






class Custom_QT5_Toolbar(NavigationToolbar2QT):
    def __init__(self, canvas, parent_pyqt5_QMainWindow, coordinates=True):
        super().__init__(canvas, parent_pyqt5_QMainWindow, coordinates)
        self.parent_pyqt5_QMainWindow = parent_pyqt5_QMainWindow
        
    def home(self, *args, **kwargs):
        # Call the Original Home Behavior. This performs the actual view reset in Matplotlib
        super().home(*args, **kwargs)
        # Now I can force my own view
        self.parent_pyqt5_QMainWindow.pcolormesh_transects.TA_image_axis_han.set_xlim(self.parent_pyqt5_QMainWindow.pcolormesh_transects.home_xlim)
        self.parent_pyqt5_QMainWindow.pcolormesh_transects.TA_image_axis_han.set_ylim(self.parent_pyqt5_QMainWindow.pcolormesh_transects.home_ylim)
        self.parent_pyqt5_QMainWindow.canvas.draw()



# Define a hook to ensure exceptions are caught by the debugger/terminal
def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    print("Uncaught exception", file=sys.stderr)
    sys.__excepthook__(exc_type, exc_value, exc_traceback)

sys.excepthook = handle_exception


# if __name__ == "__main__":       # This line is asking: "Is this file the one that started the process?"  We can have only one QApplication instance per process, so we need to ensure this is the main file.
#     # Check if a filename was provided in the command line arguments
#     if len(sys.argv) > 1:
#         TA_matrix_input = sys.argv[1]
#     else:
#         # Fallback default if no file is provided
#         TA_matrix_input = "HHHF_Zn_heme_ZnCl_p425nm_red_300uW.h5"  # default input for testing:
#     # Start the application
#     window = TA_plot_matrix_GUI(TA_matrix_input) #The QApplication instance must be created before this.  The init method recognizes QApplication already exists and uses it.
    
